/* expatParser.cpp
 A class that presents a nice C++ interface to the expat library
  (which is all plain C).  Puts its results into an object
  passed in by the user (which is subclassed from docObjectIface).
  http://projects.zillabit.com/xml.html

Copyright (c) 2000, Earl Levine 
All rights reserved.
Redistribution and use in source and binary forms, with or without modification, are permitted 
provided that the following conditions are met:

-Redistributions of source code must retain the above copyright notice, this list of conditions 
and the following disclaimer.

-The name of Earl Levine may not be used to endorse or promote products derived from this 
software without specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS'' AND ANY EXPRESS 
OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY 
AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNERS OR 
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON 
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE 
OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
POSSIBILITY OF SUCH DAMAGE. 
*/

#include "expatParser.h"
#include "docObjectIface.h"
#include <string.h>
#include "xmlparse/xmlparse.h"

// C-linkage callback functions for expat, just call expatParser methods,
//  which just call the docObjectIface's methods
extern "C" void StartElementHandler(void *userData, const XML_Char *name, const XML_Char **atts) {
	if(userData) ((expatParser*)userData)->
		StartElementHandler(name, atts);
}
void expatParser::StartElementHandler(const XML_Char *name, const XML_Char **atts) {
	if(m_docObject) m_docObject->
		StartElementHandler(name, atts);
}
extern "C" void EndElementHandler(void *userData, const XML_Char *name) {
	if(userData) ((expatParser*)userData)->
		EndElementHandler(name);
}
void expatParser::EndElementHandler(const XML_Char *name) {
	if(m_docObject) m_docObject->
		EndElementHandler(name);
}
extern "C" void CharacterDataHandler(void *userData, const XML_Char *s, int len) {
	if(userData) ((expatParser*)userData)->
		CharacterDataHandler(s, len);
}
void expatParser:: CharacterDataHandler(const XML_Char *s, int len) {
	if(m_docObject) m_docObject->
		CharacterDataHandler(s, len);
}
extern "C" void ProcessingInstructionHandler(void *userData, const XML_Char *target, const XML_Char *data) {
	if(userData) ((expatParser*)userData)->
		ProcessingInstructionHandler(target, data);
}
void expatParser::ProcessingInstructionHandler(const XML_Char *target, const XML_Char *data) {
	if(m_docObject) m_docObject->
		ProcessingInstructionHandler(target, data);
}
extern "C" void CommentHandler(void *userData, const XML_Char *data) {
	if(userData) ((expatParser*)userData)->
		CommentHandler(data);
}
void expatParser::CommentHandler(const XML_Char *data) {
	if(m_docObject) m_docObject->
		CommentHandler(data);
}
extern "C" void StartCdataSectionHandler(void *userData) {
	if(userData) ((expatParser*)userData)->
		StartCdataSectionHandler();
}
void expatParser::StartCdataSectionHandler() {
	if(m_docObject) m_docObject->
		StartCdataSectionHandler();
}
extern "C" void EndCdataSectionHandler(void *userData) {
	if(userData) ((expatParser*)userData)->
		EndCdataSectionHandler();
}
void expatParser::EndCdataSectionHandler() {
	if(m_docObject) m_docObject->
		EndCdataSectionHandler();
}
extern "C" void DefaultHandler(void *userData, const XML_Char *s, int len) {
	if(userData) ((expatParser*)userData)->
		DefaultHandler(s, len);
}
void expatParser::DefaultHandler(const XML_Char *s, int len) {
	if(m_docObject) m_docObject->
		DefaultHandler(s, len);
}
extern "C" void UnparsedEntityDeclHandler(void *userData, const XML_Char *entityName, const XML_Char *base, const XML_Char *systemId, const XML_Char *publicId, const XML_Char *notationName) {
	if(userData) ((expatParser*)userData)->
		UnparsedEntityDeclHandler(entityName, base, systemId, publicId, notationName);
}
void expatParser::UnparsedEntityDeclHandler(const XML_Char *entityName, const XML_Char *base, const XML_Char *systemId, const XML_Char *publicId, const XML_Char *notationName) {
	if(m_docObject) m_docObject->
		UnparsedEntityDeclHandler(entityName, base, systemId, publicId, notationName);
}
extern "C" void NotationDeclHandler(void *userData, const XML_Char *notationName, const XML_Char *base, const XML_Char *systemId, const XML_Char *publicId) {
	if(userData) ((expatParser*)userData)->
		NotationDeclHandler(notationName, base, systemId, publicId);
}
void expatParser::NotationDeclHandler(const XML_Char *notationName, const XML_Char *base, const XML_Char *systemId, const XML_Char *publicId) {
	if(m_docObject) m_docObject->
		NotationDeclHandler(notationName, base, systemId, publicId);
}
extern "C" void StartNamespaceDeclHandler(void *userData, const XML_Char *prefix, const XML_Char *uri) {
	if(userData) ((expatParser*)userData)->
		StartNamespaceDeclHandler(prefix, uri);
}
void expatParser::StartNamespaceDeclHandler(const XML_Char *prefix, const XML_Char *uri) {
	if(m_docObject) m_docObject->
		StartNamespaceDeclHandler(prefix, uri);
}
extern "C" void EndNamespaceDeclHandler(void *userData, const XML_Char *prefix) {
	if(userData) ((expatParser*)userData)->
		EndNamespaceDeclHandler(prefix);
}
void expatParser::EndNamespaceDeclHandler(const XML_Char *prefix) {
	if(m_docObject) m_docObject->
		EndNamespaceDeclHandler(prefix);
}
// There are other handlers declared in xmlparse.h that could be added here...

expatParser::expatParser(docObjectIface *docObject,
						 bool allowExpansionOfInternalEntities) {
	m_parserInstance = NULL;
	m_docObject = docObject;
	m_allowExpansionOfInternalEntities = allowExpansionOfInternalEntities;
}

expatParser::~expatParser() {
	if(m_parserInstance != NULL)
		XML_ParserFree(m_parserInstance);
}

bool expatParser::parse(const char *documentSegment, int documentSegmentLength, bool isFinal,
						XML_LChar **pErrorString, int *pErrorLineNumber) {
	// delete *pErrorString in case we're recycling it from an earlier call
	if(*pErrorString) delete *pErrorString;
	*pErrorString = NULL;

	if(m_parserInstance == NULL) {
		// Need to initialize the parserInstance
		m_parserInstance = XML_ParserCreate(NULL);
		if(m_parserInstance == NULL) {
			return false; // couldn't init expat
		}
		// set handlers (callbacks)
		XML_SetUserData(m_parserInstance, this);
		XML_SetElementHandler(m_parserInstance, ::StartElementHandler, ::EndElementHandler);
		XML_SetCharacterDataHandler(m_parserInstance, ::CharacterDataHandler);
		XML_SetProcessingInstructionHandler(m_parserInstance, ::ProcessingInstructionHandler);
		XML_SetCommentHandler(m_parserInstance, ::CommentHandler);
		XML_SetCdataSectionHandler(m_parserInstance, ::StartCdataSectionHandler, ::EndCdataSectionHandler);
		if(m_allowExpansionOfInternalEntities)
			XML_SetDefaultHandlerExpand(m_parserInstance, ::DefaultHandler);
		else
			XML_SetDefaultHandler(m_parserInstance, ::DefaultHandler);
		XML_SetUnparsedEntityDeclHandler(m_parserInstance, ::UnparsedEntityDeclHandler);
		XML_SetNotationDeclHandler(m_parserInstance, ::NotationDeclHandler);
		XML_SetNamespaceDeclHandler(m_parserInstance, ::StartNamespaceDeclHandler, ::EndNamespaceDeclHandler);
		// ... can add more handlers here, see xmlparse.h for full list
	}

	bool parsedOK = 
		(0!=XML_Parse(m_parserInstance, documentSegment, documentSegmentLength, isFinal));
	if(!parsedOK) { // try to get error info
		const XML_LChar* returnedErrorString 
			= XML_ErrorString(XML_GetErrorCode(m_parserInstance));
		if(returnedErrorString) {
#ifdef XML_UNICODE_WCHAR_T
			*pErrorString = new XML_LChar[1+wcslen(returnedErrorString)];
			if(*pErrorString) wcscpy(*pErrorString, returnedErrorString);
#else
			*pErrorString = new XML_LChar[1+strlen(returnedErrorString)];
			if(*pErrorString) strcpy(*pErrorString, returnedErrorString);
#endif
		}
		if(pErrorLineNumber)
			*pErrorLineNumber = XML_GetCurrentLineNumber(m_parserInstance);
	}

	if(isFinal) {
		// destroy the parserInstance now in case someone tries to call parse again
		if(m_parserInstance != NULL)
			XML_ParserFree(m_parserInstance);
		m_parserInstance = NULL;
	}

	return parsedOK;
}
